/*
 * Copyright (c) 2018 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package main

//go:generate gogitversion -p main

import (
	"flag"
	"fmt"
	"math"
	"os"
	"os/signal"
	"strings"
	"syscall"
	"time"

	"code.uplex.de/uplex-varnish/k8s-vcl-reloader/pkg/apiclient"
	"code.uplex.de/uplex-varnish/k8s-vcl-reloader/pkg/varnish"

	"github.com/sirupsen/logrus"

	api_v1 "k8s.io/api/core/v1"
	"k8s.io/client-go/informers"
	"k8s.io/client-go/kubernetes"
	"k8s.io/client-go/rest"
	"k8s.io/client-go/tools/clientcmd"
)

var (
	versionF = flag.Bool("version", false, "print version and exit")
	loglvlF  = flag.String("log-level", "INFO",
		"log level: one of PANIC, FATAL, ERROR, WARN, INFO, DEBUG, \n"+
			"or TRACE")
	namespaceF = flag.String("namespace", api_v1.NamespaceAll,
		"namespace in which to listen for resources (default all)")
	masterURLF = flag.String("masterurl", "", "cluster master URL, for "+
		"out-of-cluster runs")
	kubeconfigF = flag.String("kubeconfig", "", "config path for the "+
		"cluster master URL, for out-of-cluster runs")
	monIntvlF = flag.Duration("monitorintvl", 30*time.Second,
		"interval at which the monitor thread checks and updates\n"+
			"the Varnish instance.\n"+
			"Monitor deactivated when <= 0s")
	metricsPortF = flag.Uint("metricsport", 8080,
		"port at which to listen for the /metrics endpoint")
	resyncPeriodF = flag.Duration("resyncPeriod", 30*time.Second,
		"if non-zero, re-update the controller with the state of\n"+
			"the cluster this often, even if nothing has changed,\n"+
			"to synchronize state that may have been missed")
	admTimeoutF = flag.Duration("admTimeout", 10*time.Second,
		"timeout for connections to the Varnish instance\n"+
			"administrative port (0s for no timeout)")
	homeF = flag.String("n", "", "name of the Varnish instance,\n"+
		"the same value set for varnishd -n")
	mainF = flag.String("main", "default.vcl",
		"the VCL source to be loaded, must include all other sources")
	cfgMapF = flag.String("configmap", "",
		"k8s ConfigMap representing Varnish sources (required)")
	versionKeyF = flag.String("versionkey", "",
		"key in the ConfigMap for a VCL version string")

	logFormat = logrus.TextFormatter{
		DisableColors: true,
		FullTimestamp: true,
	}
	log = &logrus.Logger{
		Out:       os.Stdout,
		Formatter: &logFormat,
		Level:     logrus.InfoLevel,
	}
	informerStop = make(chan struct{})
)

type tmplPaths []string

func (t *tmplPaths) String() string {
	return strings.Join(*t, " ")
}
func (t *tmplPaths) Set(s string) error {
	*t = append(*t, s)
	return nil
}

type svcs []string

func (s *svcs) String() string {
	return strings.Join(*s, " ")
}
func (s *svcs) Set(svc string) error {
	*s = append(*s, svc)
	return nil
}

func main() {
	var templatePaths tmplPaths
	var services svcs

	flag.Var(&templatePaths, "template",
		"tuple of the form tmplPath:vclPath, where tmplPath is the\n"+
			"path of a template, and vclPath is the path of the\n"+
			"VCL source to be generated from the template")
	flag.Var(&services, "svc",
		"k8s Service whose Endpoints are to be configured as Varnish\n"+
			"backends")
	flag.Parse()

	if *versionF {
		fmt.Printf("%s version %s\n", os.Args[0], version)
		os.Exit(0)
	}

	lvl := strings.ToLower(*loglvlF)
	switch lvl {
	case "panic":
		log.Level = logrus.PanicLevel
	case "fatal":
		log.Level = logrus.FatalLevel
	case "error":
		log.Level = logrus.ErrorLevel
	case "warn":
		log.Level = logrus.WarnLevel
	case "debug":
		log.Level = logrus.DebugLevel
	case "trace":
		log.Level = logrus.TraceLevel
	case "info":
		break
	default:
		fmt.Printf("Unknown log level %s, exiting", *loglvlF)
		os.Exit(-1)
	}

	log.Info("Starting VCL reloader version: ", version)

	podNs, podName := os.Getenv("POD_NAMESPACE"), os.Getenv("POD_NAME")
	if podNs == "" || podName == "" {
		log.Fatalf("Env variable POD_NAMESPACE and POD_NAME required")
		os.Exit(-1)
	}
	log.Infof("Running in Pod: %s/%s", podNs, podName)

	if *cfgMapF == "" {
		log.Fatal("configmap arg is required")
		os.Exit(-1)
	}

	if *metricsPortF > math.MaxUint16 {
		log.Fatalf("metricsport %d out of range (max %d)",
			*metricsPortF, math.MaxUint16)
		os.Exit(-1)
	}

	tmplMap := make(map[string]string, len(templatePaths))
	for _, tmplPath := range templatePaths {
		tmplArgs := strings.Split(tmplPath, ":")
		if len(tmplArgs) != 2 {
			log.Fatal("illegal template arg: ", tmplPath)
			os.Exit(-1)
		}
		tmplMap[tmplArgs[0]] = tmplArgs[1]
	}

	fqSvcs := make([]string, len(services))
	for i, svc := range services {
		svcArg := strings.Split(svc, "/")
		if len(svcArg) > 2 {
			log.Fatal("illegal svc arg: ", svc)
			os.Exit(-1)
		}
		if len(svcArg) == 2 {
			if *namespaceF != api_v1.NamespaceAll &&
				svcArg[0] != *namespaceF {
				log.Fatalf("conflicting namespaces: "+
					"namespace=%s svc=%s", *namespaceF, svc)
				os.Exit(-1)
			}
			fqSvcs[i] = svc
			continue
		}
		if *namespaceF != api_v1.NamespaceAll {
			fqSvcs[i] = *namespaceF + "/" + svc
		} else {
			fqSvcs[i] = "default/" + svc
		}
	}

	var cfgMapNamespace, cfgMapName string
	cfgMapSlice := strings.Split(*cfgMapF, "/")
	if len(cfgMapSlice) > 2 {
		log.Fatal("illegal configmap arg: ", *cfgMapF)
		os.Exit(-1)
	}
	if len(cfgMapSlice) == 2 {
		if *namespaceF != api_v1.NamespaceAll &&
			cfgMapSlice[0] != *namespaceF {
			log.Fatalf("conflicting namespaces: "+
				"namespace=%s configmap=%s", *namespaceF,
				*cfgMapF)
			os.Exit(-1)
		}
		cfgMapNamespace = cfgMapSlice[0]
		cfgMapName = cfgMapSlice[1]
	} else if *namespaceF != api_v1.NamespaceAll {
		cfgMapNamespace = *namespaceF
		cfgMapName = *cfgMapF
	} else {
		cfgMapNamespace = "default"
		cfgMapName = *cfgMapF
	}

	updaterCfg := apiclient.SpecUpdaterConfig{
		SvcKeys:         fqSvcs,
		Main:            *mainF,
		CfgMapNamespace: cfgMapNamespace,
		CfgMapName:      cfgMapName,
		VersionKey:      *versionKeyF,
	}

	vController, err := varnish.NewController(log, *homeF, tmplMap, *mainF,
		*admTimeoutF, *monIntvlF)
	if err != nil {
		log.Fatal("Cannot initialize Varnish controller: ", err)
		os.Exit(-1)
	}

	var config *rest.Config
	if *masterURLF == "" && *kubeconfigF == "" {
		if config, err = rest.InClusterConfig(); err != nil {
			log.Fatalf("error creating in-cluster config: %v", err)
		}
	} else if config, err = clientcmd.BuildConfigFromFlags(*masterURLF,
		*kubeconfigF); err != nil {
		log.Fatalf("error creating client config from "+
			"masterURL=%s and kubeconfig=%s: %v",
			*masterURLF, *kubeconfigF, err)
	}
	kubeClient, err := kubernetes.NewForConfig(config)
	if err != nil {
		log.Fatal("Failed to create client:", err)
	}

	var informerFactory informers.SharedInformerFactory
	if *namespaceF == api_v1.NamespaceAll {
		informerFactory = informers.NewSharedInformerFactory(
			kubeClient, *resyncPeriodF)
	} else {
		informerFactory = informers.NewSharedInformerFactoryWithOptions(
			kubeClient, *resyncPeriodF,
			informers.WithNamespace(*namespaceF))
	}

	client, err := apiclient.NewClient(log, kubeClient, updaterCfg,
		vController, informerFactory, fqSvcs,
		cfgMapNamespace+"/"+cfgMapName, podNs, podName)
	if err != nil {
		log.Fatalf("Could not initialize API client: %v", err)
		os.Exit(-1)
	}
	vController.EvtGenerator(client)
	go handleTermination(log, client, vController)
	vController.Start()
	informerFactory.Start(informerStop)
	client.Run("", uint16(*metricsPortF))
}

func handleTermination(
	log *logrus.Logger,
	client *apiclient.Client,
	vc *varnish.Controller) {

	signalChan := make(chan os.Signal, 1)
	signal.Notify(signalChan, syscall.SIGTERM, syscall.SIGINT)

	sig := <-signalChan
	log.Infof("Received signal (%s), shutting down", sig.String())

	log.Info("Shutting down the API client")
	client.Stop()

	log.Info("Shutting down informers")
	informerStop <- struct{}{}

	log.Info("Shutting down the Varnish controller")
	vc.Quit()

	log.Info("Exiting")
	os.Exit(0)
}
