/*
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package apiclient

import (
	"fmt"
	"time"

	"code.uplex.de/uplex-varnish/k8s-vcl-reloader/pkg/varnish"
	"code.uplex.de/uplex-varnish/k8s-vcl-reloader/pkg/varnish/vcl"

	"github.com/sirupsen/logrus"

	api_v1 "k8s.io/api/core/v1"
	core_v1_listers "k8s.io/client-go/listers/core/v1"
	"k8s.io/client-go/tools/cache"
)

// SpecUpdaterConfig represents the user configuration for k8s
// resources that are considered for a VCL update.
type SpecUpdaterConfig struct {
	SvcKeys         []string
	Main            string
	CfgMapName      string
	CfgMapNamespace string
	VersionKey      string
}

// SpecUpdater encapsulates updating the VCL spec from the current
// state of the k8s cluster.
type SpecUpdater struct {
	log    *logrus.Logger
	cfg    SpecUpdaterConfig
	vc     *varnish.Controller
	svc    core_v1_listers.ServiceLister
	endp   core_v1_listers.EndpointsLister
	cfgMap core_v1_listers.ConfigMapLister
}

// NewSpecUpdater creates a SpecUpdater.
func NewSpecUpdater(
	log *logrus.Logger,
	cfg SpecUpdaterConfig,
	vc *varnish.Controller,
	listers *Listers) *SpecUpdater {

	return &SpecUpdater{
		log:    log,
		cfg:    cfg,
		vc:     vc,
		endp:   listers.endp,
		svc:    listers.svc,
		cfgMap: listers.cfgMap,
	}
}

func getEndpSpec(addr api_v1.EndpointAddress, port int32) vcl.Endpoint {
	ep := vcl.Endpoint{
		IP:   addr.IP,
		Port: port,
	}
	ep.Namespace = addr.TargetRef.Namespace
	ep.UID = string(addr.TargetRef.UID)
	ep.ResourceVersion = addr.TargetRef.ResourceVersion
	return ep
}

func (updater *SpecUpdater) getSvcSpec(
	svc *api_v1.Service) (vclSvc vcl.Service, err error) {

	updater.log.Debugf("Updating Service %s/%s spec from cluster config",
		svc.Namespace, svc.Name)

	vclSvc = vcl.Service{}
	eps, err := updater.endp.Endpoints(svc.Namespace).Get(svc.Name)
	if err != nil {
		return
	}

	vclSvc.Namespace = svc.Namespace
	vclSvc.UID = string(svc.UID)
	vclSvc.ResourceVersion = svc.ResourceVersion
	vclSvc.Endpoints = make(map[string]vcl.Endpoint)
	for _, subset := range eps.Subsets {
		if len(subset.Ports) != 1 {
			// XXX ???
			err = fmt.Errorf("Endpoint %s/%s has %d ports",
				eps.Namespace, eps.Name, len(subset.Ports))
			return
		}
		port := subset.Ports[0].Port
		for _, addr := range subset.Addresses {
			vclSvc.Endpoints[addr.TargetRef.Name] =
				getEndpSpec(addr, port)
		}
	}
	return
}

func (updater *SpecUpdater) getSpec() (vcl.Spec, error) {
	updater.log.Info("Updating VCL spec from cluster config")

	spec := vcl.Spec{
		Main:    updater.cfg.Main,
		Created: time.Now(),
	}

	svcsSpec := make(map[string]vcl.Service)
	for _, svcKey := range updater.cfg.SvcKeys {
		ns, name, err := cache.SplitMetaNamespaceKey(svcKey)
		if err != nil {
			return spec, err
		}
		svc, err := updater.svc.Services(ns).Get(name)
		if err != nil {
			return spec, err
		}
		// XXX same name in different namespaces
		svcsSpec[svc.Name], err = updater.getSvcSpec(svc)
		if err != nil {
			return spec, err
		}
	}
	if len(svcsSpec) != len(updater.cfg.SvcKeys) {
		var svcsNotFound []string
		for _, key := range updater.cfg.SvcKeys {
			_, name, err := cache.SplitMetaNamespaceKey(key)
			if err != nil {
				return spec, err
			}
			if _, exists := svcsSpec[name]; !exists {
				svcsNotFound = append(svcsNotFound, name)
			}
		}
		return spec, fmt.Errorf("Services not found, no update: %s",
			svcsNotFound)
	}
	spec.Services = svcsSpec

	cfgMap, err := updater.cfgMap.ConfigMaps(updater.cfg.CfgMapNamespace).
		Get(updater.cfg.CfgMapName)
	if err != nil {
		return spec, err
	}
	version := ""
	if updater.cfg.VersionKey != "" {
		if v, exists := cfgMap.Data[updater.cfg.VersionKey]; exists {
			version = v
		}
	}

	spec.Version = version
	spec.CfgMap.Name = cfgMap.Name
	spec.CfgMap.Namespace = cfgMap.Namespace
	spec.CfgMap.Version = version
	spec.CfgMap.UID = string(cfgMap.UID)
	spec.CfgMap.ResourceVersion = cfgMap.ResourceVersion

	return spec, nil
}

// Update generates a new configuration from the current state of the
// k8s cluster, and loads it as a new instance of VCL.
func (updater *SpecUpdater) Update() error {
	updater.log.Info("Update signal received")
	spec, err := updater.getSpec()
	if err != nil {
		// XXX generate Warn event
		updater.log.Errorf("Cannot create spec: %s", err)
		return err
	}
	if updater.vc.HasConfig(spec) {
		updater.log.Infof("Varnish VCL config is up to date")
		return nil
	}
	if err = updater.vc.Update(&spec); err != nil {
		// XXX generate Warn event, inc counter
		updater.log.Errorf("Update failed: %s", err)
		return err
	}
	// XXX generate Info event, inc counter(?)
	updater.log.Info("Varnish update succeeded")
	return nil
}
