/*
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

// Package varnish encapsulates interaction with a Varnish instance
// to transform desired states from k8s configs to the actual state of
// its VCL configuration. Only this package imports
// varnishapi/pkg/admin to interact with the CLI of the Varnish
// instance.
package varnish

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"os"
	"sync"
	"text/template"
	"time"

	"code.uplex.de/uplex-varnish/k8s-vcl-reloader/pkg/interfaces"
	"code.uplex.de/uplex-varnish/k8s-vcl-reloader/pkg/varnish/vcl"
	"code.uplex.de/uplex-varnish/varnishapi/pkg/admin"
	"code.uplex.de/uplex-varnish/varnishapi/pkg/vsm"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/sirupsen/logrus"
)

// Controller encapsulates information about the Varnish instance that
// it configures.
type Controller struct {
	log        *logrus.Logger
	updater    interfaces.Updater
	evt        interfaces.PodEventGenerator
	spec       *vcl.Spec
	cfgLoaded  bool
	home       string
	Banner     string
	tmplMap    map[string]string
	tmpl2file  map[*template.Template]string
	loadFile   string
	admTimeout time.Duration
	admMtx     *sync.Mutex
	wg         *sync.WaitGroup
	monIntvl   time.Duration
}

// NewController returns an instance of Controller.
//
//    log: logger object initialized at startup
//    home: Varnish home directory
//    tmplMap: map of template paths to file paths
//    loadFile: file to be named in a VCL load command
//    admTimeout: connection timeout for the Varnish admin port
//    monIntvl: interval at which to run the monitor
func NewController(
	log *logrus.Logger,
	home string,
	tmplMap map[string]string,
	loadFile string,
	admTimeout time.Duration,
	monIntvl time.Duration) (*Controller, error) {

	tmpl2file, err := vcl.InitTemplates(tmplMap)
	if err != nil {
		metrics.tmplFails.Inc()
		return nil, err
	}
	initMetrics()
	return &Controller{
		log:        log,
		home:       home,
		loadFile:   loadFile,
		tmplMap:    tmplMap,
		tmpl2file:  tmpl2file,
		admTimeout: admTimeout,
		monIntvl:   monIntvl,
		admMtx:     &sync.Mutex{},
		wg:         new(sync.WaitGroup),
	}, nil
}

// EvtGenerator sets the object that implements interface
// PodEventGenerator, and will be used by the monitor goroutine to
// generate Events.
func (vc *Controller) EvtGenerator(podEvt interfaces.PodEventGenerator) {
	vc.evt = podEvt
}

// SetUpdater sets the object that implements interface Updater, and
// will be used by generate a VCL spec from the current state of the
// k8s cluster.
func (vc *Controller) SetUpdater(updater interfaces.Updater) {
	vc.updater = updater
}

// Start initiates the Varnish controller and starts the monitor
// goroutine.
func (vc *Controller) Start() {
	fmt.Printf("Varnish controller logging at level: %s\n", vc.log.Level)
	go vc.monitor(vc.monIntvl)
}

func (vc *Controller) getAddrSecret() (string, []byte, error) {
	vsm := vsm.New()
	if vsm == nil {
		return "", nil, fmt.Errorf("Cannot initialize shared memory " +
			"attach (out of memory?)")
	}
	defer vsm.Destroy()

	if err := vsm.Attach(vc.home); err != nil {
		return "", nil, fmt.Errorf("Cannot attach to shared memory "+
			"for %s: %s", vc.home, vsm.Error())
	}
	addr, err := vsm.GetMgmtAddr()
	if err != nil {
		return "", nil, fmt.Errorf("Cannot get management address: %s",
			vsm.Error())
	}
	secretPath, err := vsm.GetSecretPath()
	if err != nil {
		return addr, nil, fmt.Errorf("Cannot get secret path: %s",
			vsm.Error())
	}
	file, err := os.Open(secretPath)
	if err != nil {
		return addr, nil, fmt.Errorf("Cannot open secret at %s: %s",
			secretPath, err)
	}
	secret, err := ioutil.ReadAll(file)
	if err != nil {
		return addr, nil, fmt.Errorf("Cannot read secret at %s: %s",
			secretPath, err)
	}
	return addr, secret, nil
}

// Update a Varnish Service to implement an configuration.
//
//    spec: VCL spec corresponding to the configuration
func (vc *Controller) Update(spec *vcl.Spec) error {

	var err error

	metrics.updates.Inc()
	cfgName := spec.ConfigName()
	vc.log.Infof("Update Varnish to VCL config %s: %+v", cfgName, spec)
	vc.admMtx.Lock()
	defer vc.admMtx.Unlock()
	vc.wg.Add(1)
	defer vc.wg.Done()
	vc.spec = spec
	vc.cfgLoaded = false

	vc.log.Trace("Re-parsing templates: ", vc.tmplMap)
	vc.tmpl2file, err = vcl.InitTemplates(vc.tmplMap)
	if err != nil {
		metrics.tmplFails.Inc()
		return err
	}

	vc.log.Tracef("Get address and secret from %s", vc.home)
	addr, secret, err := vc.getAddrSecret()
	if err != nil {
		metrics.vsmFails.Inc()
		return err
	}

	vc.log.Tracef("Connect to %s, timeout=%v", addr, vc.admTimeout)
	timer := prometheus.NewTimer(metrics.connectLatency)
	adm, err := admin.Dial(addr, secret, vc.admTimeout)
	timer.ObserveDuration()
	if err != nil {
		metrics.connectFails.Inc()
		return err
	}
	defer adm.Close()
	vc.Banner = adm.Banner
	vc.log.Info("Connected to Varnish admin endpoint")

	loaded := false
	vc.log.Trace("List VCLs")
	vcls, err := adm.VCLList()
	if err != nil {
		return err
	}
	vc.log.Tracef("VCL List: %+v", vcls)
	for _, vcl := range vcls {
		if vcl.Name == cfgName {
			loaded = true
			break
		}
	}
	if loaded {
		vc.log.Infof("Config %s already loaded", cfgName)
		vc.cfgLoaded = true
		return nil
	}

	vc.log.Trace("Generating VCL files from templates")
	for tmpl, file := range vc.tmpl2file {
		vc.log.Trace("Template", tmpl.Name())
		if err = vc.spec.WriteTemplate(tmpl, file); err != nil {
			metrics.templateErrs.Inc()
			return fmt.Errorf("Cannot generate %s from template "+
				"%s: %s", file, tmpl.Name(), err)
		}
	}
	vc.log.Info("VCL files generated from templates")
	metrics.templateGens.Inc()

	vc.log.Tracef("Load config %s", cfgName)
	timer = prometheus.NewTimer(metrics.vclLoadLatency)
	err = adm.VCLLoad(cfgName, vc.loadFile)
	timer.ObserveDuration()
	if err != nil {
		metrics.vclLoadErrs.Inc()
		return fmt.Errorf("Error loading config %s: %v", cfgName, err)
	}
	if err = adm.VCLUse(cfgName); err != nil {
		metrics.vclLoadErrs.Inc()
		return fmt.Errorf("Error using config %s after load: %v",
			cfgName, err)
	}
	vc.cfgLoaded = true
	metrics.vclLoads.Inc()
	vc.log.Infof("Loaded VCL config %s", cfgName)
	return nil
}

// HasConfig returns true iff a configuration is already loaded for a
// Varnish Service (so a new sync attempt is not necessary).
//
//    spec: VCL specification derived from the configuration
func (vc *Controller) HasConfig(spec vcl.Spec) bool {
	if vc.spec == nil {
		return false
	}
	if !vc.cfgLoaded {
		return false
	}
	return bytes.Equal(vc.spec.DeepHash(), spec.DeepHash())
}

// Quit stops the Varnish controller.
func (vc *Controller) Quit() {
	vc.log.Info("Wait for admin interactions with Varnish instances to " +
		"finish")
	vc.wg.Wait()
}
